@@ -1,10 +1,10 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "toml" "0.2.0"
+(define-package "toml" "0.4.0"
   "TOML (Tom's Obvious, Minimal Language) parser."
   ()
   :url "https://github.com/gongo/emacs-toml"
-  :commit "aa222ff8864f14410a042b822f59c588f6f60b05"
-  :revdesc "aa222ff8864f"
+  :commit "aea330f33f922c50e108aeca4e83159ca5da8556"
+  :revdesc "0.4.0-0-gaea330f33f92"
   :keywords '("toml" "parser")
   :authors '(("Wataru MIYAGUNI" . "gonngo@gmail.com"))
   :maintainers '(("Wataru MIYAGUNI" . "gonngo@gmail.com")))
