@@ -5,8 +5,8 @@
 ;; Author: Wataru MIYAGUNI <gonngo@gmail.com>
 ;; URL: https://github.com/gongo/emacs-toml
 ;; Keywords: toml parser
-;; Package-Version: 0.2.0
-;; Package-Revision: aa222ff8864f
+;; Package-Version: 0.4.0
+;; Package-Revision: 0.4.0-0-gaea330f33f92
 
 ;; MIT License
 ;;
@@ -41,22 +41,16 @@
 
 (require 'parse-time)
 
-(defconst toml->special-escape-characters
-  '(?b ?t ?n ?f ?r ?\" ?\/ ?\\)
-  "Characters which are escaped in TOML.
-
-\\b     - backspace       (U+0008)
-\\t     - tab             (U+0009)
-\\n     - linefeed        (U+000A)
-\\f     - form feed       (U+000C)
-\\r     - carriage return (U+000D)
-\\\"     - quote           (U+0022)
-\\\/     - slash           (U+002F)
-\\\\     - backslash       (U+005C)
-
-notes:
-
- Excluded four hex (\\uXXXX).  Do check in `toml:read-escaped-char'")
+(defconst toml->escape-sequence-alist
+  '((?b . ?\b)    ; backspace       (U+0008)
+    (?t . ?\t)    ; tab             (U+0009)
+    (?n . ?\n)    ; linefeed        (U+000A)
+    (?f . ?\f)    ; form feed       (U+000C)
+    (?r . ?\r)    ; carriage return (U+000D)
+    (?\" . ?\")   ; quote           (U+0022)
+    (?\\ . ?\\))  ; backslash       (U+005C)
+  "Alist mapping TOML escape characters to their actual values.
+Excludes \\uXXXX which is handled separately in `toml:read-escaped-char'.")
 
 (defconst toml->parse-dispatch-table
   (let ((table
@@ -68,7 +62,7 @@ notes:
            (?\' . toml:read-literal-string))))
     (mapc (lambda (char)
             (push (cons char 'toml:read-start-with-number) table))
-          '(?- ?0 ?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9))
+          '(?+ ?- ?0 ?1 ?2 ?3 ?4 ?5 ?6 ?7 ?8 ?9))
     table))
 
 (defconst toml->regexp-datetime
@@ -78,12 +72,18 @@ notes:
 \\(0[1-9]\\|[1-2][0-9]\\|3[0-1]\\)T\
 \\([0-1][0-9]\\|2[0-4]\\):\
 \\([0-5][0-9]\\):\
-\\([0-5][0-9]\\)Z"
-  "Regular expression for a datetime (Zulu time format).")
+\\([0-5][0-9]\\)\
+\\(?:\\.\\([0-9]+\\)\\)?\
+\\(Z\\|[+-][0-9]\\{2\\}:[0-9]\\{2\\}\\)"
+  "Regular expression for RFC 3339 datetime with timezone and fractional seconds.")
 
 (defconst toml->regexp-numeric
-  "\\(-?[0-9]+[\\.0-9\\]*\\)"
-  "Regular expression for a numeric.")
+  "\\([+-]?[0-9][_0-9]*[.0-9eE+_-]*\\)"
+  "Regular expression for capturing numeric-like strings.")
+
+(defconst toml->regexp-numeric-strict
+  "^[+-]?[0-9]+\\(?:_[0-9]+\\)*\\(?:\\.\\(?:[0-9]+\\(?:_[0-9]+\\)*\\)\\)?\\(?:[eE][+-]?[0-9]+\\(?:_[0-9]+\\)*\\)?$"
+  "Strict regular expression for validating numeric format.")
 
 ;; Error conditions
 
@@ -148,6 +148,10 @@ notes:
 (put 'toml-array-table-error 'error-conditions
      '(toml-array-table-error toml-error error))
 
+(put 'toml-inline-table-error 'error-message "Bad inline table")
+(put 'toml-inline-table-error 'error-conditions
+     '(toml-inline-table-error toml-error error))
+
 (defun toml:assoc (keys hash)
   "Look up nested KEYS in HASH and return the found element.
 
@@ -227,44 +231,104 @@ Move point to the end of read characters."
   (unless (eq ?\\ (toml:read-char t))
     (signal 'toml-string-escape-error (list (point))))
   (let* ((char (toml:read-char t))
-         (special (memq char toml->special-escape-characters)))
+         (mapped (assq char toml->escape-sequence-alist)))
     (cond
-     (special (concat (list ?\\ char)))
+     (mapped (char-to-string (cdr mapped)))
      ((and (eq char ?u)
            (toml:search-forward "[0-9A-Fa-f]\\{4\\}"))
-      (concat "\\u" (match-string 0)))
+      (let ((code-point (string-to-number (match-string 0) 16)))
+        (when (or (and (>= code-point #xD800) (<= code-point #xDFFF))
+                  (> code-point #x10FFFF))
+          (signal 'toml-string-unicode-escape-error (list (point))))
+        (char-to-string code-point)))
+     ((and (eq char ?U)
+           (toml:search-forward "[0-9A-Fa-f]\\{8\\}"))
+      (let ((code-point (string-to-number (match-string 0) 16)))
+        (when (or (and (>= code-point #xD800) (<= code-point #xDFFF))
+                  (> code-point #x10FFFF))
+          (signal 'toml-string-unicode-escape-error (list (point))))
+        (char-to-string code-point)))
      (t (signal 'toml-string-unicode-escape-error (list (point)))))))
 
+(defun toml:read-multiline-basic-string ()
+  "Read multiline basic string (enclosed in \"\"\") at point.
+Handles escape sequences, line-ending backslash continuation,
+and trims immediate newline after opening delimiter."
+  ;; Skip the opening """
+  (forward-char 3)
+  ;; Trim immediate newline after opening delimiter
+  (when (eq (toml:get-char-at-point) ?\n)
+    (forward-char))
+  (let ((characters '()))
+    (while (not (looking-at "\"\"\""))
+      (when (eobp)
+        (signal 'toml-string-error (list (point))))
+      (cond
+       ;; Line-ending backslash: trim the backslash, newline, and following whitespace
+       ((looking-at "\\\\[ \t]*\n[ \t\n]*")
+        (goto-char (match-end 0)))
+       (t
+        (let ((char (toml:get-char-at-point)))
+          ;; Regular escape sequence or regular character
+          (if (eq char ?\\)
+              (push (toml:read-escaped-char) characters)
+            (push (toml:read-char) characters))))))
+    ;; Skip the closing """
+    (forward-char 3)
+    (apply #'concat (nreverse characters))))
+
 (defun toml:read-string ()
   "Read string at point that surrounded by double quotation mark.
 Move point to the end of read strings."
   (unless (eq ?\" (toml:get-char-at-point))
     (signal 'toml-string-error (list (point))))
-  (let ((characters '())
-        (char (toml:read-char)))
-    (while (not (eq char ?\"))
-      (when (toml:end-of-line-p)
+  ;; Check for multiline string (""")
+  (if (looking-at "\"\"\"")
+      (toml:read-multiline-basic-string)
+    (forward-char)  ; skip opening "
+    (let ((characters '()))
+      (while (not (eq (toml:get-char-at-point) ?\"))
+        (when (toml:end-of-line-p)
+          (signal 'toml-string-error (list (point))))
+        (let ((char (toml:get-char-at-point)))
+          (if (eq char ?\\)
+              (push (toml:read-escaped-char) characters)
+            (push (toml:read-char) characters))))
+      (forward-char)  ; skip closing "
+      (apply #'concat (nreverse characters)))))
+
+(defun toml:read-multiline-literal-string ()
+  "Read multiline literal string (enclosed in ''') at point.
+No escape processing. Trims immediate newline after opening delimiter."
+  ;; Skip the opening '''
+  (forward-char 3)
+  ;; Trim immediate newline after opening delimiter
+  (when (eq (toml:get-char-at-point) ?\n)
+    (forward-char))
+  (let ((characters '()))
+    (while (not (looking-at "'''"))
+      (when (eobp)
         (signal 'toml-string-error (list (point))))
-      (push (if (eq char ?\\)
-                (toml:read-escaped-char)
-              (toml:read-char))
-            characters)
-      (setq char (toml:get-char-at-point)))
-    (forward-char)
-    (apply 'concat (nreverse characters))))
+      (push (toml:read-char) characters))
+    ;; Skip the closing '''
+    (forward-char 3)
+    (apply #'concat (nreverse characters))))
 
 (defun toml:read-literal-string ()
   "Read TOML literal string (enclosed in single quotes) at point."
   (unless (eq ?\' (toml:get-char-at-point))
     (signal 'toml-string-error (list (point))))
-  (forward-char)
-  (let ((characters '()))
-    (while (not (eq (toml:get-char-at-point) ?\'))
-      (when (toml:end-of-line-p)
-        (signal 'toml-string-error (list (point))))
-      (push (toml:read-char) characters))
+  ;; Check for multiline literal string (''')
+  (if (looking-at "'''")
+      (toml:read-multiline-literal-string)
     (forward-char)
-    (apply #'concat (nreverse characters))))
+    (let ((characters '()))
+      (while (not (eq (toml:get-char-at-point) ?\'))
+        (when (toml:end-of-line-p)
+          (signal 'toml-string-error (list (point))))
+        (push (toml:read-char) characters))
+      (forward-char)
+      (apply #'concat (nreverse characters)))))
 
 (defun toml:read-boolean ()
   "Read boolean at point.  Return t or nil.
@@ -276,31 +340,44 @@ Move point to the end of read boolean string."
     (signal 'toml-boolean-error (list (point))))))
 
 (defun toml:read-datetime ()
-  "Read datetime at point.
-Return time list (seconds, minutes, hour, day, month and year).
+  "Read RFC 3339 datetime at point.
+Return alist with keys: year, month, day, hour, minute, second,
+fraction, timezone.
 Move point to the end of read datetime string."
   (unless (toml:search-forward toml->regexp-datetime)
     (signal 'toml-datetime-error (list (point))))
-  (let ((seconds (string-to-number (match-string-no-properties 6)))
-        (minutes (string-to-number (match-string-no-properties 5)))
-        (hour    (string-to-number (match-string-no-properties 4)))
-        (day     (string-to-number (match-string-no-properties 3)))
-        (month   (string-to-number (match-string-no-properties 2)))
-        (year    (string-to-number (match-string-no-properties 1))))
-    (list seconds minutes hour day month year)))
+  (let ((year     (string-to-number (match-string-no-properties 1)))
+        (month    (string-to-number (match-string-no-properties 2)))
+        (day      (string-to-number (match-string-no-properties 3)))
+        (hour     (string-to-number (match-string-no-properties 4)))
+        (minute   (string-to-number (match-string-no-properties 5)))
+        (second   (string-to-number (match-string-no-properties 6)))
+        (fraction (match-string-no-properties 7))  ; optional
+        (timezone (match-string-no-properties 8))) ; Z or +HH:MM or -HH:MM
+    `((year . ,year)
+      (month . ,month)
+      (day . ,day)
+      (hour . ,hour)
+      (minute . ,minute)
+      (second . ,second)
+      (fraction . ,(when fraction (string-to-number (concat "0." fraction))))
+      (timezone . ,timezone))))
 
 (defun toml:read-numeric ()
   "Read numeric (integer or float) at point.  Return numeric.
 Move point to the end of read numeric string."
   (unless (toml:search-forward toml->regexp-numeric)
     (signal 'toml-numeric-error (list (point))))
-  (let ((numeric (match-string-no-properties 0)))
-    (if (with-temp-buffer
-          (insert numeric)
-          (goto-char (point-min))
-          (search-forward "." nil t 2)) ;; e.g. "0.21.1" is NG
-        (signal 'toml-numeric-error (list (point)))
-      (string-to-number numeric))))
+  (let ((numeric-str (match-string-no-properties 0)))
+    ;; Two-stage validation:
+    ;; 1. toml->regexp-numeric (loose) - greedily captures all numeric-like
+    ;;    characters to ensure invalid trailing chars (e.g., "1.1.1") are
+    ;;    included in the match rather than left unparsed.
+    ;; 2. toml->regexp-numeric-strict - validates the captured string has
+    ;;    correct format (e.g., rejects "1e", "1.", "1.1.1").
+    (unless (string-match-p toml->regexp-numeric-strict numeric-str)
+      (signal 'toml-numeric-error (list (point))))
+    (string-to-number (replace-regexp-in-string "_" "" numeric-str))))
 
 (defun toml:read-start-with-number ()
   "Read string that start with number at point.
@@ -345,8 +422,12 @@ Move point to the end of read string."
         (if (char-equal char-after-read ?,)
             (progn
               (forward-char)
-              (toml:seek-readable-point))
-          (signal 'toml-array-error (list (point))))))
+              (toml:seek-readable-point)
+              ;; TODO: Trailing comma in inline tables is allowed from TOML v1.1.0.
+              ;; Until then, it should be rejected.
+              (when (char-equal (toml:get-char-at-point) ?})
+                (signal 'toml-inline-table-error (list (point)))))
+          (signal 'toml-inline-table-error (list (point))))))
     (forward-char)
     (nreverse elements)))
 
@@ -698,11 +779,11 @@ Example:
         (if current-array-table
             ;; Add to array table element
             (setq hashes (toml:add-to-array-table current-array-table
-                                                   current-array-sub-keys
-                                                   current-key
-                                                   current-value
-                                                   hashes
-                                                   array-table-registry))
+                                                  current-array-sub-keys
+                                                  current-key
+                                                  current-value
+                                                  hashes
+                                                  array-table-registry))
           ;; Add to regular table
           (setq hashes (toml:make-table-hashes current-table
                                                current-key
@@ -716,6 +797,7 @@ Example:
   "Read the TOML object contained in STRING and return it."
   (with-temp-buffer
     (insert string)
+    (toml:--normalize-newlines)
     (goto-char (point-min))
     (toml:read)))
 
@@ -723,9 +805,20 @@ Example:
   "Read the TOML object contained in FILE and return it."
   (with-temp-buffer
     (insert-file-contents file)
+    (toml:--normalize-newlines)
     (goto-char (point-min))
     (toml:read)))
 
+(defun toml:--normalize-newlines ()
+  "Replace CRLF with LF in the current buffer.
+The TOML spec allows parsers to normalize newlines to whatever
+makes sense for their platform (see toml-lang/toml#281).
+On Unix-like systems, normalizing CRLF to LF at the entry point
+keeps the parser simple while correctly rejecting bare CR."
+  (goto-char (point-min))
+  (while (search-forward "\r\n" nil t)
+    (replace-match "\n" t t)))
+
 (provide 'toml)
 
 ;;; toml.el ends here
